#liyuezhou.cm@gmail.com
import math, cv2
import numpy as np
import torch
from vgg_backbone import build_vgg16
from feature_utils import get_subwindow_feature, generate_patch_feature, round_python2, resize_tensor, get_subwindow_image, exract_selected_deature
from siamese import SiameseNet
from image_loader import default_image_loader
from tracking_utils import fuse_feature, calculate_scale, generate_2d_window, cal_window_size
from oneshot import oneshot_model
from chanselection import  method_CAS, method_EAS,channel_select_win_rank
from aug_util import get_aug_set
class Dcda_Tracker(object):
    def __init__(self, config, model = None, display = True, device = 'cpu', name = 'DCDA', model_from_mat = True):
        """

        args:
            target_location - [x1,y1,w,h], 1-index for OTB benchmark
            TODO: model_from_mat
        """
        super(Dcda_Tracker, self).__init__()
        #---------------trackers parameters initialization--------------------------

        self.name = name
        self.config = config
        self.display = display
        self.device = device
        self.rescale = 1
        self.model_from_mat = model_from_mat

        #-------------model initialization--------------------
        if model is None:
            self.model = build_vgg16(self.config).to(self.device)
        else:
            self.model = model.to(self.device)
        self.model.train()
        self.siamese_model = SiameseNet().to(self.device)
        self.toc = 0
        self.imagenet_conv_5_1 = self.model.get_os_block().clone()
        #-------------hp settings----------------------------
        self.os_channel_num = 0.5
        self.select_weight = 1.0
        #-------------
        self.dcda_flip = {}
        self.dcda_selected={}


    def enhance_flip(self, img, target_loc):
        cv2.flip(img, 1, img)
        rescale = 1.0
        target_loc = target_loc.copy()
        target_loc[0:2] = [img.shape[1] - target_loc[0] - target_loc[2], target_loc[1]]
        target_location = target_loc
        origin_target_size = math.sqrt(target_location[2] * target_location[3])
        origin_image_size = img.shape[0: 2][::-1] # [width,height]
        if origin_target_size > self.config.MODEL.MAX_SIZE:
            rescale = self.config.MODEL.MAX_SIZE / origin_target_size
        elif origin_target_size < self.config.MODEL.MIN_SIZE:
            rescale = self.config.MODEL.MIN_SIZE / origin_target_size
        image = cv2.resize(img,tuple((np.ceil(np.array(origin_image_size) * rescale)).astype(int)),
                           interpolation=cv2.INTER_LINEAR)
        target_location = round_python2(np.array(target_location) * rescale)-np.array([1,1,0,0])#0-index
        target_size = target_location[2: 4]# [width, height]
        image_size = image.shape[0:2]# [height, width]
        search_size, ratio = cal_window_size(self.config.MODEL.MAX_SIZE, image_size, self.config.MODEL.SCALE_NUM, self.config.MODEL.TOTAL_STRIDE)
        input_size = np.array([search_size, search_size])
        srch_window_location = cal_srch_window_location(target_location, search_size)
        subwindow_image = get_subwindow_image(srch_window_location, image, input_size)
        aug_samples = get_aug_set(subwindow_image, input_size)
        aug_features = []
        for im in aug_samples:
            aug_features.append(get_subwindow_feature(self.model, im, srch_window_location, input_size))
        features = get_subwindow_feature(self.model, image, srch_window_location, input_size)
        patch_features, patch_locations = generate_patch_feature(target_size[::-1], srch_window_location, features)
        self.dcda_flip['Conv4-1'] = method_EAS(patch_features[0])
        self.dcda_flip['Conv4-3'] = method_CAS(patch_features[1], patch_features[1].shape[1], select_weight=self.select_weight)
        self.dcda_flip['Conv5-1'] = channel_select_win_rank(patch_features[2],
                select_num=int((len(self.dcda_flip['Conv4-1'])+len(self.dcda_flip['Conv4-3']))*0.5*self.os_channel_num))
        return patch_features

    def _calc_balance_weight(self, feature):
        bal_weights = []
        for i in range(len(feature)):
            bal_weights.append(torch.max(torch.sum(torch.squeeze(feature[i])[:, :, :], dim=0)))
        bal_weights = min(bal_weights) / torch.tensor(bal_weights)
        return bal_weights

    def _calc_weighted_feature(self, feature, balance_weight):
        for i in range(len(feature)):
            feature[i] = feature[i] * balance_weight[i]
        return feature

    def process_dcda_feature(self, patch_features, selection):
        feature = [exract_selected_deature(patch_features[0], selection['Conv4-1']),
                   exract_selected_deature(patch_features[1], selection['Conv4-3']),
                   exract_selected_deature(patch_features[2], selection['Conv5-1'])]
        return feature

    def process_exemplar_feature(self, patch_features):
        exemplar_flip = self.process_dcda_feature(self.flip_feature, self.dcda_selected)
        exemplar_flip = self._calc_weighted_feature(exemplar_flip, self._calc_balance_weight(exemplar_flip))

        exemplar = self.process_dcda_feature(patch_features, self.dcda_selected)
        self.balance_weights = self._calc_balance_weight(exemplar)
        exemplar = self._calc_weighted_feature(exemplar, self.balance_weights)
        exemplar_combine= [exemplar_flip[0], exemplar_flip[1], exemplar_flip[2], exemplar[0], exemplar[1], exemplar[2]]


        return exemplar_combine


    def initialize_dcda(self, img, target_loc, visualize = False):

        #------------sequence parameters initialization----------------------------
        # img = default_image_loader(img_path)#<class 'numpy.ndarray'> [height, width, channel]

        self.target_location = target_loc

        self.flip_feature = self.enhance_flip(img.copy(), target_loc.copy())

        origin_target_size = math.sqrt(self.target_location[2] * self.target_location[3])
        origin_target_location = self.target_location#<class 'list'>
        origin_image_size = img.shape[0: 2][::-1] # [width,height]
        if origin_target_size > self.config.MODEL.MAX_SIZE:
            self.rescale = self.config.MODEL.MAX_SIZE / origin_target_size
        elif origin_target_size < self.config.MODEL.MIN_SIZE:
            self.rescale = self.config.MODEL.MIN_SIZE / origin_target_size

        #----------------scale image cv2 numpy.adarray---------------
        image = cv2.resize(
                img,
                tuple((np.ceil(np.array(origin_image_size) * self.rescale)).astype(int)),
                interpolation=cv2.INTER_LINEAR
        )
        # template enhance

        #------scaled target location, get position and size [x1,y1,width,height]------
        self.target_location = round_python2(np.array(self.target_location) * self.rescale)-np.array([1,1,0,0])#0-index
        target_size = self.target_location[2: 4]# [width, height]
        image_size = image.shape[0:2]# [height, width]
        search_size, ratio = cal_window_size(self.config.MODEL.MAX_SIZE, image_size, self.config.MODEL.SCALE_NUM, self.config.MODEL.TOTAL_STRIDE)
        self.input_size = np.array([search_size, search_size])

        #------------First frame processing--------------------
        self.srch_window_location = cal_srch_window_location(self.target_location, search_size)
        features = get_subwindow_feature(self.model, image, self.srch_window_location, self.input_size, visualize = visualize)
        subwindow_image = get_subwindow_image(self.srch_window_location, image, self.input_size, visualize=False)
        # oneshot-aug
        aug_samples = get_aug_set(subwindow_image, self.input_size)
        aug_features = []
        for im in aug_samples:
            aug_features.append(get_subwindow_feature(self.model, im, self.srch_window_location, self.input_size, visualize = visualize))

        patch_features, patch_locations = generate_patch_feature(target_size[::-1], self.srch_window_location, features)
        self.feature_pad = 2
        self.b_feature_pad = int(self.feature_pad / 2)
        self.filter_sizes = [torch.tensor(feature.shape).numpy() for feature in patch_features]
        oneshot_block_weight = oneshot_model(features, self.filter_sizes, self.device, self.imagenet_conv_5_1, aug_features)
        self.model.set_os_block(oneshot_block_weight)

        features = get_subwindow_feature(self.model, image, self.srch_window_location, self.input_size, visualize = visualize)
        patch_features, patch_locations = generate_patch_feature(target_size[::-1], self.srch_window_location, features)

        # do dcda
        self.dcda_selected['Conv4-1'] = method_EAS(patch_features[0])
        self.dcda_selected['Conv4-3'] = method_CAS(patch_features[1], patch_features[1].shape[1], select_weight=self.select_weight)
        self.dcda_selected['Conv5-1'] = channel_select_win_rank(patch_features[2],
                    select_num=int((len(self.dcda_selected['Conv4-1'])+len(self.dcda_selected['Conv4-3']))*0.5*self.os_channel_num))

        self.dcda_selected['Conv4-1'] = [k for k in self.dcda_selected['Conv4-1'] if k in self.dcda_flip['Conv4-1']]
        self.dcda_selected['Conv4-3'] = [k for k in self.dcda_selected['Conv4-3'] if k in self.dcda_flip['Conv4-3']]
        self.dcda_selected['Conv5-1'] = [k for k in self.dcda_selected['Conv5-1'] if k in self.dcda_flip['Conv5-1']]

        exemplar_patch = self.process_exemplar_feature(patch_features)

        self.exemplar_features = fuse_feature(exemplar_patch)

        #------------visualization------------------------------------------------
        if self.display:
            self.visualization(img, origin_target_location, 0)
        return  origin_target_location


    def tracking(self, img, visualize = False):
        #-------------read image and rescale the image-----------------------------
        # img = default_image_loader(img_path)#<class 'numpy.ndarray'>[height, width, channel]

        image = cv2.resize(
                img,
                tuple((np.ceil(np.array(img.shape[0:2][::-1]) * self.rescale)).astype(int)),
                interpolation=cv2.INTER_LINEAR
        )
        tic = cv2.getTickCount()
        #-------------get multi-scale feature--------------------------------------
        features = get_subwindow_feature(self.model, image, self.srch_window_location, self.input_size, visualize = visualize)

        patch_features_ba = self.process_dcda_feature(features, self.dcda_selected)
        patch_features_ba = self._calc_weighted_feature(patch_features_ba, self.balance_weights)

        feature_size = (torch.tensor(patch_features_ba[0].shape)).numpy().astype(int)[-2:]
        selected_features = fuse_feature(patch_features_ba)
        selected_features = torch.cat((selected_features, selected_features), dim=1)
        selected_features_1 = resize_tensor(selected_features, tuple(feature_size + self.feature_pad))
        selected_features_3 = resize_tensor(selected_features, tuple(feature_size - self.feature_pad))
        selected_features_1 = selected_features_1[:,:,self.b_feature_pad:feature_size[0]+self.b_feature_pad, self.b_feature_pad:feature_size[1]+self.b_feature_pad]

        selected_features_3 = torch.nn.functional.pad(selected_features_3, (self.b_feature_pad, self.b_feature_pad, self.b_feature_pad, self.b_feature_pad))
        scaled_features = torch.cat((selected_features_1,selected_features,selected_features_3), dim = 0)

        #-------------get response map-----------------------------------------------
        response_map = self.siamese_model(scaled_features, self.exemplar_features).to('cpu')
        scaled_response_map = torch.squeeze(resize_tensor(
                                                response_map,
                                                tuple(self.srch_window_location[-2:].astype(int)),
                                                mode = 'bicubic',
                                                align_corners = True))
        hann_window = generate_2d_window('hann', tuple(self.srch_window_location[-2:].astype(int)), scaled_response_map.shape[0])
        scaled_response_maps = scaled_response_map + hann_window

        #-------------find max-response----------------------------------------------
        scale_ind = calculate_scale(scaled_response_maps, self.config.MODEL.SCALE_WEIGHTS)
        response_map = scaled_response_maps[scale_ind,:,:].numpy()
        max_h, max_w = np.where(response_map == np.max(response_map))
        if len(max_h)>1:
            max_h = np.array([max_h[0],])
        if len(max_w)>1:
            max_w = np.array([max_w[0],])

        #-------------update tracking state and save tracking result----------------------------------------
        target_loc_center = np.append(self.target_location[0:2]+(self.target_location[2:4])/2, self.target_location[2:4])
        target_loc_center[0:2] = target_loc_center[0:2] + (np.append(max_w,max_h)-(self.srch_window_location[2:4]/2-1))*self.config.MODEL.SCALES[scale_ind]
        target_loc_center[2:4] = target_loc_center[2:4] * self.config.MODEL.SCALES[scale_ind]

        self.target_location = np.append(target_loc_center[0:2]-(target_loc_center[2:4])/2, target_loc_center[2:4])
        self.srch_window_location[2:4] = (round_python2(self.srch_window_location[2:4] * self.config.MODEL.SCALES[scale_ind]))
        self.srch_window_location[0:2] = target_loc_center[0:2]-(self.srch_window_location[2:4])/2

        tracking_bbox = (self.target_location + np.array([1,1,0,0]))/self.rescale - np.array([1,1,0,0])#tracking_bbox: 0-index

        if self.display:
            self.visualization(img, tracking_bbox, 0)

        return tracking_bbox, np.max(response_map)/3

    def cal_fps(self, num_frame):
        toc = self.toc/cv2.getTickFrequency()
        return num_frame/toc

    def visualization(self, img, location, frame):
        show_img = img.copy()
        show_img = cv2.cvtColor(show_img, cv2.COLOR_RGB2BGR)
        bbox = [int(location[0]), int(location[1]), int(location[2]), int(location[3])]
        pt_lt = (bbox[0], bbox[1])
        pt_rb = (bbox[0] + bbox[2], bbox[1] + bbox[3])
        cv2.rectangle(show_img,pt_lt, pt_rb,(0,255,0),2)
        cv2.imshow('frame',show_img)
        cv2.waitKey(19)


def cal_srch_window_location(target_location, search_size):
    """
    function: cal_srch_window_location
    """
    srch_window_position = np.floor(target_location[0:2]+target_location[2:4]/2 - search_size/2)
    srch_window_size = np.array([search_size, search_size])
    srch_window_location = np.append(srch_window_position, srch_window_size)
    assert(srch_window_location.shape == (4,)), "the shape of srch_window_location is {}".format(srch_window_location.shape)
    return srch_window_location
